/*******************************************************************************

 Comb_cofre_dig.c

 c/ PIC 16f628a

 Senha eletrnica tipo fechadura de cofre, c/ movimentos para direita e esquerda

 Autor: Claudio Lrios

 Data: 31/07/2015    Atualizao: 01/06/2017 (incluido modo reteno na sada)

 Visualizao por meio de 2 display de 7seg (escolha catodo ou anodo comum).
 Utiliza encoder de mouse para obter valores do movimento rotativo esquerdo e
 direito.
 Desligamento do display apos 5 segundos sem acionamento.
 Senha programvel pelo usurio.
 A quantidade de nmeros da senha pode ser escolhida na compilao.
 A sada do rele poder ser programado para desligar junto com display
 (por tempo) ou  desligar quando for movimentado o encoder para qualquer um
 dos lados.
 Acrecentado opo de sada em reteno (inverte sada a cada acerto de senha)

 Programao:
 1)Aperte o boto 'prog'. Dever ligar o display e comear a piscar.
 2)Gire o encoder no sentido horrio at o valor desejado para o primeiro nmero
 3)Gire em sentido contrrio at o prximo nmero desejado.
 4)Gire em sentido horrio at o prximo nmero desejado.
 5)Mantendo esta lgica, at a quantidade de dezenas escolhidas no '#define n_dig'
 6)No ultimo numero, vire ao contrrio o encoder para grava-lo.
 7)Ir parar de piscar e acionar a sada.

 Ex. Gravar senha 87-36-05 (#define n_dig 3)
  Ao apertar o boto 'prog' os displays apresentam '00' e piscam.
 Gire sentido horrio at 87. Vire anti-horrio at 36. Vire sentido horrio at
 05. Vire sentido anti-horrio mais uma casa para gravar o ultimo nmero.

 Uso aps gravao:

 Sem apertar o boto 'prog', introduza a senha girando o encoder inicialmente
 para direita at o primeiro numero, e a partir da vire em sentido contrrio
 at chegar no segundo nmero, e assim sucessivamente at a quantidade de
 nmeros escolhidos no '#define n_dig".
 Ao entrar com o ultimo numero, vire em sentido contrario para entrar a senha.
 Ex. Usar a senha 87-36-05 (#define n_dig 3) previamente gravada.

 Gire sentido horrio at 87 (ao iniciar o movimento do encoder, ligar os
 displays). Vire anti-horrio at 36. Vire sentido horrio at 05. Vire sentido
 anti-horrio mais uma casa para entrar este ultimo nmero.

 Montagem de carater didtico apenas.
 Este arquivo  parte integrante do blog 'Larios.Tecnologia.Ws'

 /*****************************************************************************/
 //     Opes do usurio: altere e recompile com CCS C compiler
 /******************************************************************************/
   #define n_dig 3  // ajuste o nmero de dezenas que dever ter a senha (min=1)
                    // ex. #define n_dig 3  -> 05,71,93
                    //     #define n_dig 4  -> 07,97,53,21
/*****************************************************************************/
//tipo de display de 7 seguimentos usado:

//  #define display_catodo_comum

// Descomente o acima para anodo comum.
//============================================================================//

   #define tempo_display_ligado  5  // tempo do display ligado (em segundos).

   //Altere o valor acima para mais ou menos tempo .
//============================================================================//

   #define desliga_por_tempo //ao apagar display desliga tambm o rele


// comente para manter ligado o rele apos desligamento dos displays
// (desliga somente ao acionar encoder para qualquer sentido, com displays
// apagados).
//Obs.Descomente o define abaixo para poder funcionar em modo de desligamento
// por tempo.
//============================================================================//
   #define saida_retencao   //cada vez que digita a combinao inverte sada
   
//descomente para inverter a sada a cada acerto de senha
//Obs. Elimina automaticamente a funo de desligamento por tempo 
//============================================================================//

#include <16F628A.h> // pic
#use delay(clock=4000000)//frequncia do oscilador interno
#fuses NOWDT,INTRC_IO, NOPUT, NOPROTECT, NOBROWNOUT, MCLR, NOLVP, NOCPD
#use fast_io(a)// sentido das portas feito pelo programador
#use fast_io(b)//idem
#zero_ram // limpa registradores

//bits
#bit   rele    = 0x6.0  //pino 6 sada para rele
#bit   in_a    = 0x6.4 //pino 10 (rb4) entrada (Int. on change) encoder A
#bit   in_b    = 0x6.5 //pino 11 (rb5) entrada (Int. on change) encoder B
#bit   k1      = 0x6.6 //pino 12 (rb6) sada catodo 1
#bit   k2      = 0x6.7 //pino 13 (rb7) sada catodo 2
#bit   sw_prog = 0x6.3 //pino 9  (rb3) boto 'prog' senha

//bytes
#byte  porta = 0x05
#byte  portb = 0x06
#byte  trisa = 0x85
#byte  trisb = 0x86

//variveis globais
int16 time;
int8 cnt,pont,x;
int8 senha[n_dig];
int1 f_dir,f_on,f_prog,f_y,f_m;

//array
//retorna digito      abXcdefg  bits 7...0
int8 const num[]={  0b11011110 , //0
                    0b01010000 , //1
                    0b11001101 , //2
                    0b11011001 , //3
                    0b01010011 , //4
                    0b10011011 , //5
                    0b10011111 , //6
                    0b11010000 , //7
                    0b11011111 , //8
                    0b11011011  //9
                  };
//array para encoder
int8 const seqf[]={1,3,0,2};//SEQUENCIA PRA INCREMENTO
int8 const seqr[]={2,0,3,1};//SEQUENCIA PARA DECREMENTO
//constante
int16 const v_ligado= tempo_display_ligado * 125;//tempo ligado o display 7 seg

//==============================================================================
//   ROTINA DE INTERRUPO POR MUDANA DE ESTADO NOS PINOS 10 E 11 (ENCODER)
//==============================================================================

#int_rb
void isr_encoder(){
 static int8 enc_old;
 static int1 f_t1;
 int8 enc;


   //portb&=portb;
   enc=0xff; //debounce da chave
   while(--enc);
   enc=0xff;
   while(--enc);

   if(in_a){enc=1;}else {enc=0;}//atribui '1' ou '0'
   if(in_b){enc=enc+2;}//soma '0' ou '2'
   if(enc==seqf[enc_old]) {if(++cnt>199)cnt=0; if(!f_t1){f_t1=1;} else {f_dir=1;}}
   if(enc==seqr[enc_old]) {if(--cnt==0xff)cnt=199;if(f_t1){f_t1=0;} else {f_dir=0;}}
    enc_old=enc;
    f_on=1;//seta flag para ligar display
    
    time=v_ligado;//carga de tempo que ficar ligado
   }
//==============================================================================
//   ROTINA DE MULTIPLEXAO DOS 2 DISPLAY DE 7 SEGUIMENTOS (CATODOS COMUNS)
//==============================================================================


void multiplexa(){
 int8 cnt1,cnt2;
 static int1 f2;
 static int8 pisc;

   if(!f_on){ k1=0; k2=0;return;}//no est ligado
   cnt1=0;//zera contador da senha
   cnt2=cnt>>1;//divide por 2
   k2=0; k1=0;//apaga displays
   while(cnt2>9) {cnt2-=10; cnt1++;} //obtem decimal
   if(!f2){
   k1=1; f2=1;
   #ifdef display_catodo_comum
   porta=num[cnt1];
   #else
    porta=0xff^num[cnt1];
   #endif
   }
   else {
   k2=1;f2=0;
   #ifdef display_catodo_comum
   porta=num[cnt2];
   #else
    porta=0xff^num[cnt2];
   #endif
   }

   if(f_prog){
    pisc++; if(pisc>16){k1=0;k2=0;pisc+=224; } //fica piscando display 7seg
   }

}
//==============================================================================
//   ROTINA DE DESLIGAMENTO DO DISPLAY APS TEMPO
//==============================================================================

void time_on(){
 if(f_on) {//se ligado os displays
  if(!time--) {//se chegou a zero
  
   f_on=0;  f_dir=1; cnt=0; pont=0; f_prog=0; f_m=1; //reseta contadores/flags
   
  
   
   #ifndef saida_retencao
   #ifdef desliga_por_tempo
   f_y=0;
   rele=0;  //desliga a sada do rele
   #endif
   #endif
   }
  }
 }//rotina de desligamento do display
//==============================================================================
//   ROTINA DE ESCRITA/LEITURA NA EEPROM
//==============================================================================

int8 le_eeprom(int8 b){ return (read_eeprom(b));}//le eeprom
void escreve_eeprom(int8 adr,int8 v){ write_eeprom(adr,v);}//escreve eeprom

//==============================================================================
//   ROTINA DE OBTENO DA SENHA
//==============================================================================

int8 pega_senha(){

       if((!f_dir)&&(pont&1)) return(0);//ainda no pega o nmero
       if(( f_dir)&&(!(pont&1))) return(0);//ainda no pega o nmero
       if(pont<n_dig){senha[pont]= (cnt>>1);pont++; }//pega o nmero
       return (pont);//retorna valor de pont para comparao
}//pega_senha()
//==============================================================================
//   ROTINA DE GRAVAO DA SENHA
//==============================================================================

void grava_senha(){
  int8 a;
      if(f_prog){// se flag de programao setado
       for(a=0;a<n_dig;a++){
         escreve_eeprom(a,senha[a]);//grava toda senha na eeprom
         k1=0; k2=0; delay_ms(100);//apaga display
         f_prog=0;//reseta flag de programao
       }
      }
}
//==============================================================================
//   ROTINA DE COMPARAO DE SENHA  COM A GRAVADA NA EEPROM ANTERIORMENTE
//==============================================================================

int1 compara(){
   int8 a;

   for(a=0; a<n_dig;a++){
    if(senha[a]!=le_eeprom(a)) return (0);//erro, um nmero diferente da senha
    }
    return (1);//senha igual a gravada previamente na eeprom
}

void apaga_senha(){

 for(x=0;x<n_dig;x++){senha[x]=0xff;}//apaga senha inicialmente
 pont=0;
 cnt=0;
}
//==============================================================================
//   ROTINA DE MAIN
//==============================================================================

void main() {

   setup_counters(RTCC_INTERNAL,RTCC_DIV_1);//timer 0
   setup_timer_1(T1_DISABLED);//timer 1 desligado
   setup_timer_2(T2_DISABLED,0,1);//idem
   setup_comparator(NC_NC_NC_NC);//desliga comparadores interno
   setup_vref(FALSE);//desliga vref
   port_b_pullups(true);//liga pull up no portb
   enable_interrupts(int_rb);//liga interrupo por mudana de estado (rb4...rb7)
   enable_interrupts(GLOBAL);//liga geral das interrupes
   trisa=0b0000000;//sentido da porta
   trisb=0b00111110;//sentido da portb
   f_dir=1;//situao inicial
   rele=0;//desliga sada do rele
   cnt=0;//zera contador
   pont=0;//zera ponteiro
   f_m=0;
   f_y=1;
   apaga_senha();//zera senha
//==============================================================================
//   LAO PRINCIPAL DE REPETIO
//==============================================================================

   for(;;){
     #ifndef saida_retencao
     #ifndef desliga_por_tempo
     if((f_on)&&(f_y)&& f_m){ apaga_senha(); rele=0;f_y=0;f_m=0;f_on=0;}
     #endif
     #endif
     if(!sw_prog){f_prog=1; time=v_ligado; f_y=0; f_on=1;apaga_senha(); }//carga de tempo que ficar ligado
     multiplexa();//multiplexa os displays de 7 seguimento
     time_on();//tempo em que display ficar ligado
     delay_ms(8);//delay
     if(pega_senha()==(n_dig)){//compara senha ou grava se boto prog estiver apertado
     grava_senha();
     
     #ifndef  saida_retencao
     if(compara()){ rele=1; f_y=1;f_m=0;}
  //   else {rele=0;}//aciona se senha for correta
     #else
      if(compara()){rele=!rele; apaga_senha(); }     
     #endif

     }
   }
}
//            Fim do programa
